@@ -1,4 +1,4 @@
-# Copyright 2010-2020 The pygit2 contributors
+# Copyright 2010-2023 The pygit2 contributors
 #
 # This file is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2,
@@ -29,8 +29,7 @@ This is an special module, it provides stuff used by by pygit2 at run-time.
 
 # Import from the Standard Library
 import codecs
-import os
-from os.path import abspath, dirname
+from pathlib import Path
 import sys
 
 # Import from cffi
@@ -45,11 +44,12 @@ except ImportError:
 
 # C_HEADER_SRC
 if getattr(sys, 'frozen', False):
-    dir_path = getattr(sys, '_MEIPASS', None)
-    if dir_path is None:
-        dir_path = dirname(abspath(sys.executable))
+    if hasattr(sys, '_MEIPASS'):
+        dir_path = Path(sys._MEIPASS)
+    else:
+        dir_path = Path(sys.executable).parent
 else:
-    dir_path = dirname(abspath(__file__))
+    dir_path = Path(__file__).parent.absolute()
 
 # Order matters
 h_files = [
@@ -61,10 +61,10 @@ h_files = [
     'strarray.h',
     'diff.h',
     'checkout.h',
-    'pack.h',
     'transport.h',
     'proxy.h',
     'indexer.h',
+    'pack.h',
     'remote.h',
     'clone.h',
     'common.h',
@@ -77,30 +77,28 @@ h_files = [
     'net.h',
     'refspec.h',
     'repository.h',
+    'commit.h',
     'revert.h',
     'stash.h',
     'submodule.h',
+    'callbacks.h', # Bridge from libgit2 to Python
 ]
 h_source = []
 for h_file in h_files:
-    h_file = os.path.join(dir_path, 'decl', h_file)
+    h_file = dir_path / 'decl' / h_file
     with codecs.open(h_file, 'r', 'utf-8') as f:
         h_source.append(f.read())
 
 C_HEADER_SRC = '\n'.join(h_source)
 
-# Preamble
-preamble = "#include <git2.h>"
-
-# C_KEYWORDS
-libgit2_bin, libgit2_include, libgit2_lib = get_libgit2_paths()
-C_KEYWORDS = dict(libraries=['git2'],
-                  library_dirs=[libgit2_lib],
-                  include_dirs=[libgit2_include])
-
 # ffi
+_, libgit2_kw = get_libgit2_paths()
 ffi = FFI()
-ffi.set_source("pygit2._libgit2", preamble, **C_KEYWORDS)
+ffi.set_source(
+    "pygit2._libgit2",
+    "#include <git2.h>", # preamble
+    **libgit2_kw
+)
 ffi.cdef(C_HEADER_SRC)
 
 
