@@ -1,4 +1,4 @@
-# Copyright 2010-2021 The pygit2 contributors
+# Copyright 2010-2023 The pygit2 contributors
 #
 # This file is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2,
@@ -28,19 +28,22 @@ from io import BytesIO
 from string import hexdigits
 import tarfile
 from time import time
+import typing
 import warnings
 
 # Import from pygit2
 from ._pygit2 import Repository as _Repository, init_file_backend
 from ._pygit2 import Oid, GIT_OID_HEXSZ, GIT_OID_MINPREFIXLEN
-from ._pygit2 import GIT_CHECKOUT_SAFE, GIT_CHECKOUT_RECREATE_MISSING, GIT_DIFF_NORMAL
+from ._pygit2 import GIT_DIFF_NORMAL
 from ._pygit2 import GIT_FILEMODE_LINK
 from ._pygit2 import GIT_BRANCH_LOCAL, GIT_BRANCH_REMOTE, GIT_BRANCH_ALL
 from ._pygit2 import GIT_REF_SYMBOLIC
-from ._pygit2 import Reference, Tree, Commit, Blob
+from ._pygit2 import GIT_REFERENCES_ALL, GIT_REFERENCES_BRANCHES, GIT_REFERENCES_TAGS
+from ._pygit2 import GIT_CHECKOUT_SAFE, GIT_CHECKOUT_RECREATE_MISSING
+from ._pygit2 import Reference, Tree, Commit, Blob, Signature
 from ._pygit2 import InvalidSpecError
 
-from .callbacks import git_fetch_options
+from .callbacks import git_checkout_options, git_fetch_options, git_stash_apply_options
 from .config import Config
 from .errors import check_error
 from .ffi import ffi, C
@@ -86,20 +89,20 @@ class BaseRepository(_Repository):
 
     def pack(self, path=None, pack_delegate=None, n_threads=None):
         """Pack the objects in the odb chosen by the pack_delegate function
-        and write .pack and .idx files for them.
+        and write `.pack` and `.idx` files for them.
 
         Returns: the number of objects written to the pack
 
         Parameters:
 
         path
-            The path to which the .pack and .idx files should be written. None will write to the default location.
+            The path to which the `.pack` and `.idx` files should be written. `None` will write to the default location.
 
         pack_delegate
             The method which will provide add the objects to the pack builder. Defaults to all objects.
 
         n_threads
-            The number of threads the PackBuilder will spawn. If set to 0 libgit2 will autodetect the number of CPUs.
+            The number of threads the `PackBuilder` will spawn. If set to 0, libgit2 will autodetect the number of CPUs.
         """
 
         def pack_all_objects(pack_builder):
@@ -134,7 +137,7 @@ class BaseRepository(_Repository):
             The path within the parent repository to add the submodule
 
         link
-            Should workdir contain a gitlink to the repo in .git/modules vs. repo directly in workdir.
+            Should workdir contain a gitlink to the repo in `.git/modules` vs. repo directly in workdir.
         """
         csub = ffi.new('git_submodule **')
         curl = ffi.new('char[]', to_bytes(url))
@@ -148,7 +151,7 @@ class BaseRepository(_Repository):
 
         # prepare options
         opts = ffi.new('git_submodule_update_options *')
-        C.git_submodule_update_init_options(opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
+        C.git_submodule_update_options_init(opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
 
         with git_fetch_options(callbacks, opts=opts.fetch_opts) as payload:
             crepo = ffi.new('git_repository **')
@@ -164,7 +167,7 @@ class BaseRepository(_Repository):
 
     def lookup_submodule(self, path):
         """
-        Lookup submodule information by name or path.
+        Look up submodule information by name or path.
         """
         csub = ffi.new('git_submodule **')
         cpath = ffi.new('char[]', to_bytes(path))
@@ -186,7 +189,7 @@ class BaseRepository(_Repository):
 
         # prepare options
         opts = ffi.new('git_submodule_update_options *')
-        C.git_submodule_update_init_options(opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
+        C.git_submodule_update_options_init(opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
 
         with git_fetch_options(callbacks, opts=opts.fetch_opts) as payload:
             i = 1 if init else 0
@@ -214,7 +217,7 @@ class BaseRepository(_Repository):
         return self.git_object_lookup_prefix(key) is not None
 
     def __repr__(self):
-        return "pygit2.Repository(%r)" % self.path
+        return f"pygit2.Repository({repr(self.path)})"
 
     #
     # Remotes
@@ -291,6 +294,16 @@ class BaseRepository(_Repository):
         return self.create_reference_symbolic(name, target, force,
                                               message=message)
 
+    def listall_references(self) -> typing.List[str]:
+        """Return a list with all the references in the repository.
+        """
+        return list(x.name for x in self.references.iterator())
+
+    def listall_reference_objects(self) -> typing.List[Reference]:
+        """Return a list with all the reference objects in the repository.
+        """
+        return list(x for x in self.references.iterator())
+
     def resolve_refish(self, refish):
         """Convert a reference-like short name "ref-ish" to a valid
         (commit, reference) pair.
@@ -318,59 +331,35 @@ class BaseRepository(_Repository):
     #
     # Checkout
     #
-    @staticmethod
-    def _checkout_args_to_options(strategy=None, directory=None, paths=None):
-        # Create the options struct to pass
-        copts = ffi.new('git_checkout_options *')
-        check_error(C.git_checkout_init_options(copts, 1))
-
-        # References we need to keep to strings and so forth
-        refs = []
-
-        # pygit2's default is SAFE | RECREATE_MISSING
-        copts.checkout_strategy = GIT_CHECKOUT_SAFE | GIT_CHECKOUT_RECREATE_MISSING
-        # and go through the arguments to see what the user wanted
-        if strategy:
-            copts.checkout_strategy = strategy
-
-        if directory:
-            target_dir = ffi.new('char[]', to_bytes(directory))
-            refs.append(target_dir)
-            copts.target_directory = target_dir
-
-        if paths:
-            strarray = StrArray(paths)
-            refs.append(strarray)
-            copts.paths = strarray.array[0]
-
-        return copts, refs
 
     def checkout_head(self, **kwargs):
         """Checkout HEAD
 
         For arguments, see Repository.checkout().
         """
-        copts, refs = Repository._checkout_args_to_options(**kwargs)
-        check_error(C.git_checkout_head(self._repo, copts))
+        with git_checkout_options(**kwargs) as payload:
+            err = C.git_checkout_head(self._repo, payload.checkout_options)
+            payload.check_error(err)
 
     def checkout_index(self, index=None, **kwargs):
         """Checkout the given index or the repository's index
 
         For arguments, see Repository.checkout().
         """
-        copts, refs = Repository._checkout_args_to_options(**kwargs)
-        check_error(C.git_checkout_index(self._repo, index._index if index else ffi.NULL, copts))
+        with git_checkout_options(**kwargs) as payload:
+            err = C.git_checkout_index(self._repo, index._index if index else ffi.NULL, payload.checkout_options)
+            payload.check_error(err)
 
     def checkout_tree(self, treeish, **kwargs):
         """Checkout the given treeish
 
         For arguments, see Repository.checkout().
         """
-        copts, refs = Repository._checkout_args_to_options(**kwargs)
-        cptr = ffi.new('git_object **')
-        ffi.buffer(cptr)[:] = treeish._pointer[:]
-
-        check_error(C.git_checkout_tree(self._repo, cptr[0], copts))
+        with git_checkout_options(**kwargs) as payload:
+            cptr = ffi.new('git_object **')
+            ffi.buffer(cptr)[:] = treeish._pointer[:]
+            err = C.git_checkout_tree(self._repo, cptr[0], payload.checkout_options)
+            payload.check_error(err)
 
     def checkout(self, refname=None, **kwargs):
         """
@@ -397,6 +386,15 @@ class BaseRepository(_Repository):
             A list of files to checkout from the given reference.
             If paths is provided, HEAD will not be set to the reference.
 
+        callbacks : CheckoutCallbacks
+            Optional. Supply a `callbacks` object to get information about
+            conflicted files, updated files, etc. as the checkout is being
+            performed. The callbacks can also abort the checkout prematurely.
+
+            The callbacks should be an object which inherits from
+            `pyclass:CheckoutCallbacks`. It should implement the callbacks
+            as overridden methods.
+
         Examples:
 
         * To checkout from the HEAD, just pass 'HEAD'::
@@ -474,7 +472,7 @@ class BaseRepository(_Repository):
             try:
                 obj = obj.peel(Tree)
             except Exception:
-                raise TypeError('unexpected "%s"' % type(obj))
+                raise TypeError(f'unexpected "{type(obj)}"')
 
         return obj
 
@@ -567,6 +565,14 @@ class BaseRepository(_Repository):
 
         raise ValueError("Only blobs and treeish can be diffed")
 
+    def state(self) -> int:
+        """Determines the state of a git repository - ie, whether an operation
+        (merge, cherry-pick, etc) is in progress.
+        
+        Returns a GIT_REPOSITORY_STATE_* constant.
+        """
+        return C.git_repository_state(self._repo)
+
     def state_cleanup(self):
         """Remove all the metadata associated with an ongoing command like
         merge, revert, cherry-pick, etc. For example: MERGE_HEAD, MERGE_MSG,
@@ -613,7 +619,7 @@ class BaseRepository(_Repository):
         """
 
         options = ffi.new('git_blame_options *')
-        C.git_blame_init_options(options, C.GIT_BLAME_OPTIONS_VERSION)
+        C.git_blame_options_init(options, C.GIT_BLAME_OPTIONS_VERSION)
         if flags:
             options.flags = flags
         if min_match_characters:
@@ -694,18 +700,21 @@ class BaseRepository(_Repository):
         for k, v in merged_dict.items():
             enum = mapping.get(k, None)
             if enum is None:
-                raise ValueError("unknown %s: %s" % (label, k))
+                raise ValueError(f"unknown {label}: {k}")
             if v:
                 bitmask |= enum
         return bitmask
 
     @classmethod
-    def _merge_options(cls, favor='normal', flags={}, file_flags={}):
-        """Return a 'git_merge_opts *'"""
+    def _merge_options(cls, favor='normal', flags=None, file_flags=None):
+        """Return a 'git_merge_opts *'
+        """
+        flags = flags or {}
+        file_flags = file_flags or {}
 
         favor_val = cls._FAVOR_TO_ENUM.get(favor, None)
         if favor_val is None:
-            raise ValueError("unknown favor: %s" % favor)
+            raise ValueError(f"unknown favor: {favor}")
 
         flags_bitmask = Repository._flag_dict_to_bitmask(
             flags,
@@ -721,7 +730,7 @@ class BaseRepository(_Repository):
         )
 
         opts = ffi.new('git_merge_options *')
-        err = C.git_merge_init_options(opts, C.GIT_MERGE_OPTIONS_VERSION)
+        err = C.git_merge_options_init(opts, C.GIT_MERGE_OPTIONS_VERSION)
         check_error(err)
 
         opts.file_favor = favor_val
@@ -763,7 +772,7 @@ class BaseRepository(_Repository):
 
         return ret
 
-    def merge_commits(self, ours, theirs, favor='normal', flags={}, file_flags={}):
+    def merge_commits(self, ours, theirs, favor='normal', flags=None, file_flags=None):
         """
         Merge two arbitrary commits.
 
@@ -816,6 +825,8 @@ class BaseRepository(_Repository):
         Both "ours" and "theirs" can be any object which peels to a commit or
         the id (string or Oid) of an object which peels to a commit.
         """
+        flags = flags or {}
+        file_flags = file_flags or {}
 
         ours_ptr = ffi.new('git_commit **')
         theirs_ptr = ffi.new('git_commit **')
@@ -839,7 +850,7 @@ class BaseRepository(_Repository):
 
         return Index.from_c(self, cindex)
 
-    def merge_trees(self, ancestor, ours, theirs, favor='normal', flags={}, file_flags={}):
+    def merge_trees(self, ancestor, ours, theirs, favor='normal', flags=None, file_flags=None):
         """
         Merge two trees.
 
@@ -892,6 +903,8 @@ class BaseRepository(_Repository):
             * patience. Use the "patience diff" algorithm
             * minimal. Take extra time to find minimal diff
         """
+        flags = flags or {}
+        file_flags = file_flags or {}
 
         ancestor_ptr = ffi.new('git_tree **')
         ours_ptr = ffi.new('git_tree **')
@@ -920,6 +933,77 @@ class BaseRepository(_Repository):
 
         return Index.from_c(self, cindex)
 
+    def merge(self, id, favor="normal", flags=None, file_flags=None):
+        """
+        Merges the given id into HEAD.
+
+        Merges the given commit(s) into HEAD, writing the results into the working directory.
+        Any changes are staged for commit and any conflicts are written to the index.
+        Callers should inspect the repository's index after this completes,
+        resolve any conflicts and prepare a commit.
+
+        Parameters:
+
+        id
+            The id to merge into HEAD
+
+        favor
+            How to deal with file-level conflicts. Can be one of
+
+            * normal (default). Conflicts will be preserved.
+            * ours. The "ours" side of the conflict region is used.
+            * theirs. The "theirs" side of the conflict region is used.
+            * union. Unique lines from each side will be used.
+
+            For all but NORMAL, the index will not record a conflict.
+
+        flags
+            A dict of str: bool to turn on or off functionality while merging.
+            If a key is not present, the default will be used. The keys are:
+
+            * find_renames. Detect file renames. Defaults to True.
+            * fail_on_conflict. If a conflict occurs, exit immediately instead
+              of attempting to continue resolving conflicts.
+            * skip_reuc. Do not write the REUC extension on the generated index.
+            * no_recursive. If the commits being merged have multiple merge
+              bases, do not build a recursive merge base (by merging the
+              multiple merge bases), instead simply use the first base.
+
+        file_flags
+            A dict of str: bool to turn on or off functionality while merging.
+            If a key is not present, the default will be used. The keys are:
+
+            * standard_style. Create standard conflicted merge files.
+            * diff3_style. Create diff3-style file.
+            * simplify_alnum. Condense non-alphanumeric regions for simplified
+              diff file.
+            * ignore_whitespace. Ignore all whitespace.
+            * ignore_whitespace_change. Ignore changes in amount of whitespace.
+            * ignore_whitespace_eol. Ignore whitespace at end of line.
+            * patience. Use the "patience diff" algorithm
+            * minimal. Take extra time to find minimal diff
+        """
+        if not isinstance(id, (str, Oid)):
+            raise TypeError(f'expected oid (string or <Oid>) got {type(id)}')
+
+        id = self[id].id
+        c_id = ffi.new("git_oid *")
+        ffi.buffer(c_id)[:] = id.raw[:]
+
+        merge_opts = self._merge_options(favor, flags=flags or {}, file_flags=file_flags or {})
+
+        checkout_opts = ffi.new("git_checkout_options *")
+        C.git_checkout_options_init(checkout_opts, 1)
+        checkout_opts.checkout_strategy = GIT_CHECKOUT_SAFE | GIT_CHECKOUT_RECREATE_MISSING
+
+        commit_ptr = ffi.new("git_annotated_commit **")
+        err = C.git_annotated_commit_lookup(commit_ptr, self._repo, c_id)
+        check_error(err)
+
+        err = C.git_merge(self._repo, commit_ptr, 1, merge_opts, checkout_opts)
+        C.git_annotated_commit_free(commit_ptr[0])
+        check_error(err)
+
     #
     # Describe
     #
@@ -983,7 +1067,7 @@ class BaseRepository(_Repository):
         """
 
         options = ffi.new('git_describe_options *')
-        C.git_describe_init_options(options, C.GIT_DESCRIBE_OPTIONS_VERSION)
+        C.git_describe_options_init(options, C.GIT_DESCRIBE_OPTIONS_VERSION)
 
         if max_candidates_tags is not None:
             options.max_candidates_tags = max_candidates_tags
@@ -1043,8 +1127,15 @@ class BaseRepository(_Repository):
     #
     # Stash
     #
-    def stash(self, stasher, message=None, keep_index=False,
-              include_untracked=False, include_ignored=False):
+    def stash(
+            self,
+            stasher: Signature,
+            message: typing.Optional[str] = None,
+            keep_index: bool = False,
+            include_untracked: bool = False,
+            include_ignored: bool = False,
+            keep_all: bool = False,
+            paths: typing.Optional[typing.List[str]] = None):
         """
         Save changes to the working directory to the stash.
 
@@ -1067,44 +1158,47 @@ class BaseRepository(_Repository):
         include_ignored : bool
             Also stash ignored files.
 
+        keep_all : bool
+            All changes in the index and working directory are left intact.
+
+        paths : list[str]
+            An optional list of paths that control which files are stashed.
+
         Example::
 
             >>> repo = pygit2.Repository('.')
             >>> repo.stash(repo.default_signature(), 'WIP: stashing')
         """
 
-        if message:
-            stash_msg = ffi.new('char[]', to_bytes(message))
-        else:
-            stash_msg = ffi.NULL
+        opts = ffi.new('git_stash_save_options *')
+        C.git_stash_save_options_init(opts, C.GIT_STASH_SAVE_OPTIONS_VERSION)
 
         flags = 0
         flags |= keep_index * C.GIT_STASH_KEEP_INDEX
+        flags |= keep_all * C.GIT_STASH_KEEP_ALL
         flags |= include_untracked * C.GIT_STASH_INCLUDE_UNTRACKED
         flags |= include_ignored * C.GIT_STASH_INCLUDE_IGNORED
+        opts.flags = flags
 
         stasher_cptr = ffi.new('git_signature **')
         ffi.buffer(stasher_cptr)[:] = stasher._pointer[:]
+        opts.stasher = stasher_cptr[0]
+
+        if message:
+            message_ref = ffi.new('char[]', to_bytes(message))
+            opts.message = message_ref
+
+        if paths:
+            arr = StrArray(paths)
+            opts.paths = arr.array[0]
 
         coid = ffi.new('git_oid *')
-        err = C.git_stash_save(coid, self._repo, stasher_cptr[0], stash_msg, flags)
+        err = C.git_stash_save_with_opts(coid, self._repo, opts)
+
         check_error(err)
 
         return Oid(raw=bytes(ffi.buffer(coid)[:]))
 
-    @staticmethod
-    def _stash_args_to_options(reinstate_index=False, **kwargs):
-        stash_opts = ffi.new('git_stash_apply_options *')
-        check_error(C.git_stash_apply_init_options(stash_opts, 1))
-
-        flags = reinstate_index * C.GIT_STASH_APPLY_REINSTATE_INDEX
-        stash_opts.flags = flags
-
-        copts, refs = Repository._checkout_args_to_options(**kwargs)
-        stash_opts.checkout_options = copts[0]
-
-        return stash_opts
-
     def stash_apply(self, index=0, **kwargs):
         """
         Apply a stashed state in the stash list to the working directory.
@@ -1118,6 +1212,18 @@ class BaseRepository(_Repository):
         reinstate_index : bool
             Try to reinstate stashed changes to the index.
 
+        callbacks : StashApplyCallbacks
+            Optional. Supply a `callbacks` object to get information about
+            the progress of the stash application as it is being performed.
+
+            The callbacks should be an object which inherits from
+            `pyclass:StashApplyCallbacks`. It should implement the callbacks
+            as overridden methods.
+
+            Note that this class inherits from CheckoutCallbacks, so you can
+            also get information from the checkout part of the unstashing
+            process via the callbacks.
+
         The checkout options may be customized using the same arguments taken by
         Repository.checkout().
 
@@ -1127,8 +1233,9 @@ class BaseRepository(_Repository):
             >>> repo.stash(repo.default_signature(), 'WIP: stashing')
             >>> repo.stash_apply(strategy=GIT_CHECKOUT_ALLOW_CONFLICTS)
         """
-        stash_opts = Repository._stash_args_to_options(**kwargs)
-        check_error(C.git_stash_apply(self._repo, index, stash_opts))
+        with git_stash_apply_options(**kwargs) as payload:
+            err = C.git_stash_apply(self._repo, index, payload.stash_apply_options)
+            payload.check_error(err)
 
     def stash_drop(self, index=0):
         """
@@ -1147,8 +1254,9 @@ class BaseRepository(_Repository):
 
         For arguments, see Repository.stash_apply().
         """
-        stash_opts = Repository._stash_args_to_options(**kwargs)
-        check_error(C.git_stash_pop(self._repo, index, stash_opts))
+        with git_stash_apply_options(**kwargs) as payload:
+            err = C.git_stash_pop(self._repo, index, payload.stash_apply_options)
+            payload.check_error(err)
 
     #
     # Utility for writing a tree into an archive
@@ -1180,18 +1288,17 @@ class BaseRepository(_Repository):
         Example::
 
             >>> import tarfile, pygit2
-            >>>> with tarfile.open('foo.tar', 'w') as archive:
-            >>>>     repo = pygit2.Repository('.')
-            >>>>     repo.write_archive(repo.head.target, archive)
+            >>> with tarfile.open('foo.tar', 'w') as archive:
+            >>>     repo = pygit2.Repository('.')
+            >>>     repo.write_archive(repo.head.target, archive)
         """
 
         # Try to get a tree form whatever we got
-        if isinstance(treeish, Tree):
-            tree = treeish
-
         if isinstance(treeish, (str, Oid)):
             treeish = self[treeish]
 
+        tree = treeish.peel(Tree)
+
         # if we don't have a timestamp, try to get it from a commit
         if not timestamp:
             try:
@@ -1204,8 +1311,6 @@ class BaseRepository(_Repository):
         if not timestamp:
             timestamp = int(time())
 
-        tree = treeish.peel(Tree)
-
         index = Index()
         index.read_tree(tree)
 
@@ -1222,7 +1327,7 @@ class BaseRepository(_Repository):
                 info.size = 0
                 archive.addfile(info)
             else:
-                info.mode = tree[entry.path].filemode
+                info.mode = entry.mode
                 archive.addfile(info, BytesIO(content))
 
     #
@@ -1233,8 +1338,8 @@ class BaseRepository(_Repository):
         Calculate how many different commits are in the non-common parts of the
         history between the two given ids.
 
-        Ahead is how many commits are in the ancestry of the 'local' commit
-        which are not in the 'upstream' commit. Behind is the opposite.
+        Ahead is how many commits are in the ancestry of the `local` commit
+        which are not in the `upstream` commit. Behind is the opposite.
 
         Returns: a tuple of two integers with the number of commits ahead and
         behind respectively.
@@ -1270,7 +1375,7 @@ class BaseRepository(_Repository):
         """
         Retrieve an attribute for a file by path.
 
-        Returns: a boolean, None if the value is unspecified, or string with
+        Returns: a boolean, `None` if the value is unspecified, or string with
         the value of the attribute.
 
         Parameters:
@@ -1283,8 +1388,17 @@ class BaseRepository(_Repository):
             The name of the attribute to look up.
 
         flags
-            A combination of GIT_ATTR_CHECK_ flags which determine the
+            A combination of `GIT_ATTR_CHECK_` flags which determine the
             lookup order.
+
+        Examples::
+
+            >>> print(repo.get_attr('splash.bmp', 'binary'))
+            True
+            >>> print(repo.get_attr('splash.bmp', 'unknown-attr'))
+            None
+            >>> repo.get_attr('test.h', 'whitespace')
+            'tab-in-indent,trailing-space'
         """
 
         cvalue = ffi.new('char **')
@@ -1318,7 +1432,7 @@ class BaseRepository(_Repository):
         return (ffi.string(cname).decode('utf-8'), ffi.string(cemail).decode('utf-8'))
 
     def set_ident(self, name, email):
-        """Set the identity to be used for reference operations
+        """Set the identity to be used for reference operations.
 
         Updates to some references also append data to their
         reflog. You can use this method to set what identity will be
@@ -1330,7 +1444,7 @@ class BaseRepository(_Repository):
 
     def revert_commit(self, revert_commit, our_commit, mainline=0):
         """
-        Reverts the given Commit against the given "our" Commit, producing an
+        Revert the given Commit against the given "our" Commit, producing an
         Index that reflects the result of the revert.
 
         Returns: an Index with the result of the revert.
@@ -1354,7 +1468,7 @@ class BaseRepository(_Repository):
         ffi.buffer(our_commit_ptr)[:] = our_commit._pointer[:]
 
         opts = ffi.new('git_merge_options *')
-        err = C.git_merge_init_options(opts, C.GIT_MERGE_OPTIONS_VERSION)
+        err = C.git_merge_options_init(opts, C.GIT_MERGE_OPTIONS_VERSION)
         check_error(err)
 
         err = C.git_revert_commit(
@@ -1364,10 +1478,123 @@ class BaseRepository(_Repository):
 
         return Index.from_c(self, cindex)
 
+    #
+    # Amend commit
+    #
+    def amend_commit(self, commit, refname, author=None,
+                     committer=None, message=None, tree=None,
+                     encoding='UTF-8'):
+        """
+        Amend an existing commit by replacing only explicitly passed values,
+        return the rewritten commit's oid.
+
+        This creates a new commit that is exactly the same as the old commit,
+        except that any explicitly passed values will be updated. The new
+        commit has the same parents as the old commit.
+
+        You may omit the `author`, `committer`, `message`, `tree`, and
+        `encoding` parameters, in which case this will use the values
+        from the original `commit`.
+
+        Parameters:
+
+        commit : Commit, Oid, or str
+            The commit to amend.
+
+        refname : Reference or str
+            If not `None`, name of the reference that will be updated to point
+            to the newly rewritten commit. Use "HEAD" to update the HEAD of the
+            current branch and make it point to the rewritten commit.
+            If you want to amend a commit that is not currently the tip of the
+            branch and then rewrite the following commits to reach a ref, pass
+            this as `None` and update the rest of the commit chain and ref
+            separately.
+
+        author : Signature
+            If not None, replace the old commit's author signature with this
+            one.
+
+        committer : Signature
+            If not None, replace the old commit's committer signature with this
+            one.
+
+        message : str
+            If not None, replace the old commit's message with this one.
+
+        tree : Tree, Oid, or str
+            If not None, replace the old commit's tree with this one.
+
+        encoding : str
+            Optional encoding for `message`.
+        """
+
+        # Initialize parameters to pass on to C function git_commit_amend.
+        # Note: the pointers are all initialized to NULL by default.
+        coid = ffi.new('git_oid *')
+        commit_cptr = ffi.new('git_commit **')
+        refname_cstr = ffi.NULL
+        author_cptr = ffi.new('git_signature **')
+        committer_cptr = ffi.new('git_signature **')
+        message_cstr = ffi.NULL
+        encoding_cstr = ffi.NULL
+        tree_cptr = ffi.new('git_tree **')
+
+        # Get commit as pointer to git_commit.
+        if isinstance(commit, (str, Oid)):
+            commit = self[commit]
+        elif isinstance(commit, Commit):
+            pass
+        elif commit is None:
+            raise ValueError("the commit to amend cannot be None")
+        else:
+            raise TypeError("the commit to amend must be a Commit, str, or Oid")
+        commit = commit.peel(Commit)
+        ffi.buffer(commit_cptr)[:] = commit._pointer[:]
+
+        # Get refname as C string.
+        if isinstance(refname, Reference):
+            refname_cstr = ffi.new('char[]', to_bytes(refname.name))
+        elif type(refname) is str:
+            refname_cstr = ffi.new('char[]', to_bytes(refname))
+        elif refname is not None:
+            raise TypeError("refname must be a str or Reference")
+
+        # Get author as pointer to git_signature.
+        if isinstance(author, Signature):
+            ffi.buffer(author_cptr)[:] = author._pointer[:]
+        elif author is not None:
+            raise TypeError("author must be a Signature")
+
+        # Get committer as pointer to git_signature.
+        if isinstance(committer, Signature):
+            ffi.buffer(committer_cptr)[:] = committer._pointer[:]
+        elif committer is not None:
+            raise TypeError("committer must be a Signature")
+
+        # Get message and encoding as C strings.
+        if message is not None:
+            message_cstr = ffi.new('char[]', to_bytes(message, encoding))
+            encoding_cstr = ffi.new('char[]', to_bytes(encoding))
+
+        # Get tree as pointer to git_tree.
+        if tree is not None:
+            if isinstance(tree, (str, Oid)):
+                tree = self[tree]
+            tree = tree.peel(Tree)
+            ffi.buffer(tree_cptr)[:] = tree._pointer[:]
+
+        # Amend the commit.
+        err = C.git_commit_amend(coid, commit_cptr[0], refname_cstr,
+                                 author_cptr[0], committer_cptr[0],
+                                 encoding_cstr, message_cstr, tree_cptr[0])
+        check_error(err)
+
+        return Oid(raw=bytes(ffi.buffer(coid)[:]))
+
 
 class Branches:
 
-    def __init__(self, repository, flag=GIT_BRANCH_ALL, commit=None):
+    def __init__(self, repository: BaseRepository, flag: int = GIT_BRANCH_ALL, commit=None):
         self._repository = repository
         self._flag = flag
         if commit is not None:
@@ -1381,7 +1608,7 @@ class Branches:
             self.local = Branches(repository, flag=GIT_BRANCH_LOCAL, commit=commit)
             self.remote = Branches(repository, flag=GIT_BRANCH_REMOTE, commit=commit)
 
-    def __getitem__(self, name):
+    def __getitem__(self, name: str):
         branch = None
         if self._flag & GIT_BRANCH_LOCAL:
             branch = self._repository.lookup_branch(name, GIT_BRANCH_LOCAL)
@@ -1390,11 +1617,11 @@ class Branches:
             branch = self._repository.lookup_branch(name, GIT_BRANCH_REMOTE)
 
         if branch is None or not self._valid(branch):
-            raise KeyError('Branch not found: {}'.format(name))
+            raise KeyError(f'Branch not found: {name}')
 
         return branch
 
-    def get(self, key):
+    def get(self, key: str):
         try:
             return self[key]
         except KeyError:
@@ -1405,10 +1632,10 @@ class Branches:
             if self._commit is None or self.get(branch_name) is not None:
                 yield branch_name
 
-    def create(self, name, commit, force=False):
+    def create(self, name: str, commit, force=False):
         return self._repository.create_branch(name, commit, force)
 
-    def delete(self, name):
+    def delete(self, name: str):
         self[name].delete()
 
     def _valid(self, branch):
@@ -1431,29 +1658,63 @@ class Branches:
 
 class References:
 
-    def __init__(self, repository):
+    def __init__(self, repository: BaseRepository):
         self._repository = repository
 
-    def __getitem__(self, name):
+    def __getitem__(self, name: str):
         return self._repository.lookup_reference(name)
 
-    def get(self, key):
+    def get(self, key: str):
         try:
             return self[key]
         except KeyError:
             return None
 
     def __iter__(self):
-        for ref_name in self._repository.listall_references():
-            yield ref_name
+        iter = self._repository.references_iterator_init()
+        while True:
+            ref = self._repository.references_iterator_next(iter)
+            if ref:
+                yield ref.name
+            else:
+                return
+
+    def iterator(self, references_return_type:int = GIT_REFERENCES_ALL):
+        """ Creates a new iterator and fetches references for a given repository.
+
+        Can also filter and pass all refs or only branches or only tags.
+
+        Parameters:
+
+        references_return_type: int
+            Optional specifier to filter references. By default, all references are
+            returned.
+
+            The following values are accepted:
+            0 -> GIT_REFERENCES_ALL, fetches all refs, this is the default
+            1 -> GIT_REFERENCES_BRANCHES, fetches only branches
+            2 -> GIT_REFERENCES_TAGS, fetches only tags
+
+        TODO: Add support for filtering by reference types notes and remotes.
+        """
+
+        if references_return_type not in (GIT_REFERENCES_ALL, GIT_REFERENCES_BRANCHES, GIT_REFERENCES_TAGS):
+            raise ValueError("Parameter references_return_type is invalid")
+        iter = self._repository.references_iterator_init()
+        while True:
+            ref = self._repository.references_iterator_next(iter, references_return_type)
+            if ref:
+                yield ref
+            else:
+                return
 
     def create(self, name, target, force=False):
         return self._repository.create_reference(name, target, force)
 
-    def delete(self, name):
+    def delete(self, name: str):
         self[name].delete()
 
-    def __contains__(self, name):
+    def __contains__(self, name: str):
         return self.get(name) is not None
 
     @property
